freeze;

//  Based on 
//  Cassels, J.W.S.: Rational quadratic forms 
//  London ; New York : Academic Press, 1978.


intrinsic IsIsotropic5(
  a1::RngIntElt, a2::RngIntElt, a3::RngIntElt, a4::RngIntElt, a5::RngIntElt : 
  Al := "CremonaRusin") -> BoolElt, ModTupRngElt
{ Returns whether the quadratic form is isotropic over rationals 
 and isotropic vector with integral coordinates if so }

  // solve trivial case: zero coefficient occures 
  if (a1 eq 0) then
    return true, Vector([1,0,0,0,0]);
  elif (a2 eq 0) then
    return true, Vector([0,1,0,0,0]);
  elif (a3 eq 0) then
    return true, Vector([0,0,1,0,0]);
  elif (a4 eq 0) then
    return true, Vector([0,0,0,1,0]);
  elif (a5 eq 0) then
    return true, Vector([0,0,0,0,1]);
  end if;

  //  test solvability:
  //  is equivalent to the solvability over reals
  if ((a1 gt 0) and (a2 gt 0) and (a3 gt 0) and (a4 gt 0) and (a5 gt 0)) or 
     ((a1 lt 0) and (a2 lt 0) and (a3 lt 0) and (a4 lt 0) and (a5 lt 0)) then 
    return false, _; 
  end if;

  //  find t such that
  //    f1(X) = a1*x1^2 + a2*x2^2 - t*u^2 is isotropic and
  //    f2(X) = a3*x3^2 + a4*x4^2 + a5*x5^2 + t*v^2 is isotropic

  //  if t = a1*b1^2 + a2*b2^2 then f1 is isotropic ...
  //  generate all numbers representable by a1*x1^2 + a2*x2^2 and test 
  //  if f2 with the generated t is isoptropic 

  solution := Vector([0,0,0,0,0]);
  sum := 1;
  found := false;

  while (not found) do
    for b1 := 0 to sum do

      b2 := sum - b1;
      t := a1*b1^2 + a2*b2^2;

      if (t ne 0) then
        b, v := IsIsotropic4(a3,a4,a5,t : Al:=Al);
        if b then
          // construct the solution
          solution := Vector([b1*v[4], b2*v[4], v[1], v[2], v[3]]);
          found := true;
          break b1;
        end if;
      else
        solution := Vector([b1, b2, 0, 0, 0]);
        found := true;
        break b1;
      end if;
    end for;

    sum := sum+1;
  end while;

  return true, solution;
end intrinsic;


intrinsic IsIsotropic5(QF_M::AlgMatElt :
  Al := "CremonaRusin") -> BoolElt, ModTupRngElt
{ Returns whether the quadratic form is isotropic over rationals 
 and isotropic vector with integral coordinates if so }

  require IsSymmetric(QF_M): "Symmetric matrix required.";
  require NumberOfRows(QF_M) eq 5: "5 x 5 matrix required.";

  QF_M1, T, rank := OrthogonalizeGram(QF_M);

  b,v := IsIsotropic5(QF_M1[1][1], QF_M1[2][2], QF_M1[3][3], QF_M1[4][4], QF_M1[5][5] : Al := Al);
  if b then
    return true, v * T;
  end if;

  return false, _;

end intrinsic;


intrinsic IsIsotropic(a::ModTupRngElt : 
  Al := "CremonaRusin") -> BoolElt, ModTupRngElt
{ Returns whether the quadratic form is isotropic over rationals 
 and isotropic vector with integral coordinates if so }

  length := Ncols(a);
  require (length ge 2): "Too short vector.";
  if (length eq 2) then return IsIsotropic2(a[1], a[2]);
  elif (length eq 3) then return IsIsotropic3(a[1], a[2], a[3] : Al := Al);
  elif (length eq 4) then return IsIsotropic4(a[1], a[2], a[3], a[4] : Al := Al);
  elif (length eq 5) then return IsIsotropic5(a[1], a[2], a[3], a[4], a[5] : Al := Al);
  end if;

  V := RSpace(IntegerRing(), length);
  solution := V ! 0;

  // solve trivial case: zero coefficient occures 
  for i:=1 to length do
    if (a[i] eq 0) then
      solution[i] := 1;
      return true, solution;
    end if;
  end for;

  //  test solvability:
  //  is equivalent to the solvability over reals
  sign := Sign(a[3]);
  for i:=4 to length do
    if (Sign(a[i]) ne sign) then
      sign := 0;
      break i;
    end if;
  end for;

  if (sign ne 0) then
    if (Sign(a[1]) eq sign) and (Sign(a[2]) eq sign) then 
      return false, _;
    end if; 
  end if;

  //  find t such that
  //    f1(X) = a1*x1^2 + a2*x2^2 - t*u^2 is isotropic and
  //    f2(X) = a3*x3^2 + .. + a_k*x_k^2 + t*v^2 is isotropic

  //  if t = a1*b1^2 + a2*b2^2 then f1 is isotropic 
  //  f2 is isotropic if it is isotropic over reals ...
  //  generate numbers representable by a1*x1^2 + a2*x2^2 and 
  //  test posibly test if the signs in f2 are not all the same 

  b1 := 0;
  b2 := 0;

  sum := 1;
  found := false;

  while (not found) do
    b1 := -1;
    while not found and (b1 le sum) do
      b1 := b1+1;
      b2 := sum - b1;
      t := a[1]*b1^2 + a[2]*b2^2;

      if (sign eq 0) or (Sign(t) ne sign) then
        found := true;
      end if;
    end while;

    sum := sum+1;
  end while;

  if (t eq 0) then
    solution[1] := b1;
    solution[2] := b2;
  else
    if (length eq 6) then
      bool, v := IsIsotropic5(a[3], a[4], a[5], a[6], t : Al := Al);
    else
      b := RSpace(IntegerRing(), length-1) ! 0;
      for i:=3 to length do
        b[i-2] := a[i];
      end for;
      b[length-1] := t;

      bool, v := IsIsotropic(b : Al := Al);
    end if;
    solution[1] := b1 * v[length-1];
    solution[2] := b2 * v[length-1];
    for i:=3 to length do
      solution[i] := v[i-2];
    end for;
  end if;

  return true, solution;
end intrinsic;


intrinsic IsIsotropic(QF_M::AlgMatElt :
  Al := "CremonaRusin") -> BoolElt, ModTupRngElt
{ Returns whether the quadratic form is isotropic over rationals 
 and isotropic vector with integral coordinates if so }

  require IsSymmetric(QF_M): "Symmetric matrix required.";
  n := NumberOfRows(QF_M);
  require (n ge 2): "At least 2x2 matrix required.";

  QF_M1, T, rank := OrthogonalizeGram(QF_M);
  b := false;
  v := RSpace(Integers(), n) ! 0;

  if (n eq 2) then
    b,v := IsIsotropic2(QF_M1[1][1], QF_M1[2][2]);
  elif (n eq 3) then
    b,v := IsIsotropic3(QF_M1[1][1], QF_M1[2][2], QF_M1[3][3] : Al := Al);
  elif (n eq 4) then
    b,v := IsIsotropic4(QF_M1[1][1], QF_M1[2][2], QF_M1[3][3], QF_M1[4][4] : Al := Al);
  elif (n eq 5) then
    b,v := IsIsotropic5(QF_M1[1][1], QF_M1[2][2], QF_M1[3][3], QF_M1[4][4], QF_M1[5][5] : Al := Al);
  else
    a := RSpace(IntegerRing(), n) ! 0;
    for i:=1 to n do
      a[i] := QF_M1[i][i];
    end for;
  b,v := IsIsotropic(a : Al := Al);
  end if;

  if b then return true, v*T;
  else return false, _;
  end if;
end intrinsic;


